@@ -10,18 +10,25 @@ use nom::{AsChar, Err, IResult};
 use std::borrow::Cow;
 use unicode_categories::UnicodeCategories;
 
-pub(crate) fn tokenize(mut input: &str) -> Vec<Token<'_>> {
+pub(crate) fn tokenize(mut input: &str, named_placeholders: bool) -> Vec<Token<'_>> {
     let mut tokens: Vec<Token> = Vec::new();
 
+    let mut last_reserved_token = None;
+    let mut last_reserved_top_level_token = None;
+
     // Keep processing the string until it is empty
     while let Ok(result) = get_next_token(
         input,
         tokens.last().cloned(),
-        tokens
-            .iter()
-            .rfind(|token| token.kind == TokenKind::Reserved)
-            .cloned(),
+        last_reserved_token.clone(),
+        last_reserved_top_level_token.clone(),
+        named_placeholders,
     ) {
+        if result.1.kind == TokenKind::Reserved {
+            last_reserved_token = Some(result.1.clone());
+        } else if result.1.kind == TokenKind::ReservedTopLevel {
+            last_reserved_top_level_token = Some(result.1.clone());
+        }
         input = result.0;
 
         tokens.push(result.1);
@@ -83,15 +90,24 @@ fn get_next_token<'a>(
     input: &'a str,
     previous_token: Option<Token<'a>>,
     last_reserved_token: Option<Token<'a>>,
+    last_reserved_top_level_token: Option<Token<'a>>,
+    named_placeholders: bool,
 ) -> IResult<&'a str, Token<'a>> {
     get_whitespace_token(input)
         .or_else(|_| get_comment_token(input))
         .or_else(|_| get_string_token(input))
         .or_else(|_| get_open_paren_token(input))
         .or_else(|_| get_close_paren_token(input))
-        .or_else(|_| get_placeholder_token(input))
         .or_else(|_| get_number_token(input))
-        .or_else(|_| get_reserved_word_token(input, previous_token, last_reserved_token))
+        .or_else(|_| {
+            get_reserved_word_token(
+                input,
+                previous_token,
+                last_reserved_token,
+                last_reserved_top_level_token,
+            )
+        })
+        .or_else(|_| get_placeholder_token(input, named_placeholders))
         .or_else(|_| get_word_token(input))
         .or_else(|_| get_operator_token(input))
 }
@@ -208,6 +224,11 @@ fn get_string_token(input: &str) -> IResult<&str, Token<'_>> {
             take_till_escaping('\'', &['\'', '\\']),
             take(1usize),
         ))),
+        recognize(tuple((
+            tag("E'"),
+            take_till_escaping('\'', &['\'', '\\']),
+            take(1usize),
+        ))),
     ))(input)
     .map(|(input, token)| {
         (
@@ -288,12 +309,23 @@ fn get_close_paren_token(input: &str) -> IResult<&str, Token<'_>> {
     })
 }
 
-fn get_placeholder_token(input: &str) -> IResult<&str, Token<'_>> {
-    alt((
-        get_ident_named_placeholder_token,
-        get_string_named_placeholder_token,
-        get_indexed_placeholder_token,
-    ))(input)
+fn get_placeholder_token(input: &str, named_placeholders: bool) -> IResult<&str, Token<'_>> {
+    // The precedence changes based on 'named_placeholders' but not the exhaustiveness.
+    // This is to ensure the formatting is the same even if parameters aren't used.
+
+    if named_placeholders {
+        alt((
+            get_ident_named_placeholder_token,
+            get_string_named_placeholder_token,
+            get_indexed_placeholder_token,
+        ))(input)
+    } else {
+        alt((
+            get_indexed_placeholder_token,
+            get_ident_named_placeholder_token,
+            get_string_named_placeholder_token,
+        ))(input)
+    }
 }
 
 fn get_indexed_placeholder_token(input: &str) -> IResult<&str, Token<'_>> {
@@ -327,7 +359,7 @@ fn get_indexed_placeholder_token(input: &str) -> IResult<&str, Token<'_>> {
 
 fn get_ident_named_placeholder_token(input: &str) -> IResult<&str, Token<'_>> {
     recognize(tuple((
-        alt((char('@'), char(':'))),
+        alt((char('@'), char(':'), char('$'))),
         take_while1(|item: char| {
             item.is_alphanumeric() || item == '.' || item == '_' || item == '$'
         }),
@@ -393,7 +425,7 @@ fn scientific_notation(input: &str) -> IResult<&str, &str> {
     recognize(tuple((
         alt((decimal_number, digit1)),
         tag("e"),
-        alt((tag("-"), tag("+"))),
+        alt((tag("-"), tag("+"), tag(""))),
         digit1,
     )))(input)
 }
@@ -402,6 +434,7 @@ fn get_reserved_word_token<'a>(
     input: &'a str,
     previous_token: Option<Token<'a>>,
     last_reserved_token: Option<Token<'a>>,
+    last_reserved_top_level_token: Option<Token<'a>>,
 ) -> IResult<&'a str, Token<'a>> {
     // A reserved word cannot be preceded by a "."
     // this makes it so in "my_table.from", "from" is not considered a reserved word
@@ -412,7 +445,7 @@ fn get_reserved_word_token<'a>(
     }
 
     alt((
-        get_top_level_reserved_token,
+        get_top_level_reserved_token(last_reserved_top_level_token),
         get_newline_reserved_token(last_reserved_token),
         get_top_level_reserved_token_no_indent,
         get_plain_reserved_token,
@@ -429,49 +462,64 @@ fn get_uc_words(input: &str, words: usize) -> String {
         .to_ascii_uppercase()
 }
 
-fn get_top_level_reserved_token(input: &str) -> IResult<&str, Token<'_>> {
-    let uc_input = get_uc_words(input, 3);
-    let result: IResult<&str, &str> = alt((
-        terminated(tag("ADD"), end_of_word),
-        terminated(tag("AFTER"), end_of_word),
-        terminated(tag("ALTER COLUMN"), end_of_word),
-        terminated(tag("ALTER TABLE"), end_of_word),
-        terminated(tag("DELETE FROM"), end_of_word),
-        terminated(tag("EXCEPT"), end_of_word),
-        terminated(tag("FETCH FIRST"), end_of_word),
-        terminated(tag("FROM"), end_of_word),
-        terminated(tag("GROUP BY"), end_of_word),
-        terminated(tag("GO"), end_of_word),
-        terminated(tag("HAVING"), end_of_word),
-        terminated(tag("INSERT INTO"), end_of_word),
-        terminated(tag("INSERT"), end_of_word),
-        terminated(tag("LIMIT"), end_of_word),
-        terminated(tag("MODIFY"), end_of_word),
-        terminated(tag("ORDER BY"), end_of_word),
-        terminated(tag("SELECT"), end_of_word),
-        terminated(tag("SET CURRENT SCHEMA"), end_of_word),
-        terminated(tag("SET SCHEMA"), end_of_word),
-        terminated(tag("SET"), end_of_word),
-        alt((
-            terminated(tag("UPDATE"), end_of_word),
-            terminated(tag("VALUES"), end_of_word),
-            terminated(tag("WHERE"), end_of_word),
-        )),
-    ))(&uc_input);
-    if let Ok((_, token)) = result {
-        let final_word = token.split(' ').last().unwrap();
-        let input_end_pos = input.to_ascii_uppercase().find(final_word).unwrap() + final_word.len();
-        let (token, input) = input.split_at(input_end_pos);
-        Ok((
-            input,
-            Token {
-                kind: TokenKind::ReservedTopLevel,
-                value: token,
-                key: None,
-            },
-        ))
-    } else {
-        Err(Err::Error(Error::new(input, ErrorKind::Alt)))
+fn get_top_level_reserved_token<'a>(
+    last_reserved_top_level_token: Option<Token<'a>>,
+) -> impl FnMut(&'a str) -> IResult<&'a str, Token<'a>> {
+    move |input: &'a str| {
+        let uc_input: String = get_uc_words(input, 3);
+        let result: IResult<&str, &str> = alt((
+            terminated(tag("ADD"), end_of_word),
+            terminated(tag("AFTER"), end_of_word),
+            terminated(tag("ALTER COLUMN"), end_of_word),
+            terminated(tag("ALTER TABLE"), end_of_word),
+            terminated(tag("DELETE FROM"), end_of_word),
+            terminated(tag("EXCEPT"), end_of_word),
+            terminated(tag("FETCH FIRST"), end_of_word),
+            terminated(tag("FROM"), end_of_word),
+            terminated(tag("GROUP BY"), end_of_word),
+            terminated(tag("GO"), end_of_word),
+            terminated(tag("HAVING"), end_of_word),
+            terminated(tag("INSERT INTO"), end_of_word),
+            terminated(tag("INSERT"), end_of_word),
+            terminated(tag("LIMIT"), end_of_word),
+            terminated(tag("MODIFY"), end_of_word),
+            terminated(tag("ORDER BY"), end_of_word),
+            terminated(tag("SELECT"), end_of_word),
+            terminated(tag("SET CURRENT SCHEMA"), end_of_word),
+            terminated(tag("SET SCHEMA"), end_of_word),
+            terminated(tag("SET"), end_of_word),
+            alt((
+                terminated(tag("UPDATE"), end_of_word),
+                terminated(tag("VALUES"), end_of_word),
+                terminated(tag("WHERE"), end_of_word),
+                terminated(tag("RETURNING"), end_of_word),
+            )),
+        ))(&uc_input);
+        if let Ok((_, token)) = result {
+            let final_word = token.split(' ').last().unwrap();
+            let input_end_pos =
+                input.to_ascii_uppercase().find(final_word).unwrap() + final_word.len();
+            let (token, input) = input.split_at(input_end_pos);
+            let kind = if token == "EXCEPT"
+                && last_reserved_top_level_token.is_some()
+                && last_reserved_top_level_token.as_ref().unwrap().value == "SELECT"
+            {
+                // If the query statement before and after the except keyword is not complete, mark it a `Word`
+                TokenKind::Word
+            } else {
+                TokenKind::ReservedTopLevel
+            };
+            Ok((
+                input,
+                Token {
+                    kind,
+                    value: token,
+                    key: None,
+                },
+            ))
+        } else {
+            Err(Err::Error(Error::new(input, ErrorKind::Alt)))
+        }
     }
 }
 
@@ -528,11 +576,14 @@ fn get_newline_reserved_token<'a>(
 fn get_top_level_reserved_token_no_indent(input: &str) -> IResult<&str, Token<'_>> {
     let uc_input = get_uc_words(input, 2);
     let result: IResult<&str, &str> = alt((
+        terminated(tag("BEGIN"), end_of_word),
+        terminated(tag("DECLARE"), end_of_word),
         terminated(tag("INTERSECT"), end_of_word),
         terminated(tag("INTERSECT ALL"), end_of_word),
         terminated(tag("MINUS"), end_of_word),
         terminated(tag("UNION"), end_of_word),
         terminated(tag("UNION ALL"), end_of_word),
+        terminated(tag("$$"), end_of_word),
     ))(&uc_input);
     if let Ok((_, token)) = result {
         let final_word = token.split(' ').last().unwrap();
@@ -550,8 +601,10 @@ fn get_top_level_reserved_token_no_indent(input: &str) -> IResult<&str, Token<'_
         Err(Err::Error(Error::new(input, ErrorKind::Alt)))
     }
 }
-
 fn get_plain_reserved_token(input: &str) -> IResult<&str, Token<'_>> {
+    alt((get_plain_reserved_two_token, get_plain_reserved_one_token))(input)
+}
+fn get_plain_reserved_one_token(input: &str) -> IResult<&str, Token<'_>> {
     let uc_input = get_uc_words(input, 1);
     let result: IResult<&str, &str> = alt((
         terminated(tag("ACCESSIBLE"), end_of_word),
@@ -568,7 +621,6 @@ fn get_plain_reserved_token(input: &str) -> IResult<&str, Token<'_>> {
         terminated(tag("AUTOCOMMIT"), end_of_word),
         terminated(tag("AUTO_INCREMENT"), end_of_word),
         terminated(tag("BACKUP"), end_of_word),
-        terminated(tag("BEGIN"), end_of_word),
         terminated(tag("BETWEEN"), end_of_word),
         terminated(tag("BINLOG"), end_of_word),
         terminated(tag("BOTH"), end_of_word),
@@ -577,7 +629,6 @@ fn get_plain_reserved_token(input: &str) -> IResult<&str, Token<'_>> {
         alt((
             terminated(tag("CHANGE"), end_of_word),
             terminated(tag("CHANGED"), end_of_word),
-            terminated(tag("CHARACTER SET"), end_of_word),
             terminated(tag("CHARSET"), end_of_word),
             terminated(tag("CHECK"), end_of_word),
             terminated(tag("CHECKSUM"), end_of_word),
@@ -718,8 +769,6 @@ fn get_plain_reserved_token(input: &str) -> IResult<&str, Token<'_>> {
                                     terminated(tag("NOW()"), end_of_word),
                                     terminated(tag("NULL"), end_of_word),
                                     terminated(tag("OFFSET"), end_of_word),
-                                    terminated(tag("ON DELETE"), end_of_word),
-                                    terminated(tag("ON UPDATE"), end_of_word),
                                     alt((
                                         terminated(tag("ON"), end_of_word),
                                         terminated(tag("ONLY"), end_of_word),
@@ -911,6 +960,30 @@ fn get_plain_reserved_token(input: &str) -> IResult<&str, Token<'_>> {
     }
 }
 
+fn get_plain_reserved_two_token(input: &str) -> IResult<&str, Token<'_>> {
+    let uc_input = get_uc_words(input, 2);
+    let result: IResult<&str, &str> = alt((
+        terminated(tag("CHARACTER SET"), end_of_word),
+        terminated(tag("ON DELETE"), end_of_word),
+        terminated(tag("ON UPDATE"), end_of_word),
+    ))(&uc_input);
+    if let Ok((_, token)) = result {
+        let final_word = token.split(' ').last().unwrap();
+        let input_end_pos = input.to_ascii_uppercase().find(final_word).unwrap() + final_word.len();
+        let (token, input) = input.split_at(input_end_pos);
+        Ok((
+            input,
+            Token {
+                kind: TokenKind::Reserved,
+                value: token,
+                key: None,
+            },
+        ))
+    } else {
+        Err(Err::Error(Error::new(input, ErrorKind::Alt)))
+    }
+}
+
 fn get_word_token(input: &str) -> IResult<&str, Token<'_>> {
     take_while1(is_word_character)(input).map(|(input, token)| {
         (
